#include "renderer.hpp"

#include "shape.hpp"
#include "shape_visitor.hpp"
#include "circle.hpp"
#include "rectangle.hpp"

#include <iostream>

namespace shapes {
	namespace { // Alle in diesem Block hat internal linkage, ist also privat für diese CPP-Datei.
	
		// Eine Hilfsfunktion um das Color enum auszugeben.
		void print_color(Color color) {
			switch(color) {
				case Color::red:
				std::cout << "red";
				break;
				case Color::green:
				std::cout << "green";
				break;
				case Color::blue:
				std::cout << "blue";
				break;
			}
		}
	
		// Unsere konkrete Implementierung des Renderers.
		// Leitet public von Renderer ab, um die Schnittstelle zu implementieren.
		// Leitet private von Shape_visitor ab um als Visitor agieren zu können. Das muss aber nicht public
		//   sein, weil wir den Visitor nur in der Klasse selbst verwenden.
		class Cmd_renderer final : public Renderer, private Shape_visitor {
		  public:
			void draw(const Shape& shape) override {
				// Zeichnet die Shape über das Visitor-Pattern (Double-Dispatch).
				// Die visit-Implementierungen von Circle und Rectangle rufen unsere visit-Methoden unten
				//   mit ihrem tatsächlichen dynamischen Typ auf.
				shape.visit(*this);
			}
			
			// Überschreiben der visit-Methoden aus Shape_visitor, die das eigentliche Zeichnen/Ausgeben erledigen.
			void visit(const Circle& circle) override {
				std::cout << "Draw ";
				print_color(circle.color());
				std::cout << " R" << circle.radius();
				std::cout << " Circle at " << circle.x() << "/" << circle.y() << "\n";
			}
			void visit(const Rectangle& rect) override {
				std::cout << "Draw ";
				print_color(rect.color());
				std::cout << " " << rect.width() << "x" << rect.height();
				std::cout << " Rectangle at " << rect.x() << "/" << rect.y() << "\n";
			}
		};
		
	}
	
	// In der Implementierung unserer Factory-Funktion können wir die private (lokal linkage) Implementierung
	//   für unseren Renderer sehen und erzeugen.
	// Der gesammte restliche Code im Projekt sieht aber weiterhin nur die Renderer-Schnittstelle
	//   und den std::unique_ptr, der das Heap-Objekt mit dem Cmd_renderer verwaltet.
	std::unique_ptr<Renderer> create_renderer() {
		return std::make_unique<Cmd_renderer>();
	}

} // namespace shapes
